package generated.javacard.copycardMoneySecure.copycard;

import javacard.framework.ISO7816;
import javacard.framework.ISOException;
import javacard.framework.Util;

/**
 * Class for T(L)V encoding of messages and other classes
 */
public class Coding {

	/**
	 * Die maximale Laenge des Datenblocks einer APDU
	 */
	//private final static short MAX_CODEABLE_LENGTH = 256;

	/**
	 * Zeigt auf die Stelle von encodeDestination, an die geschrieben / von der gelesen werden soll
	 */
	private short encoding_length;

	/**
	 * In dieses Array wird hineinkodiert oder auch herauskodiert
	 */
	private byte[] encodeDestination;

	/**
	 * Die Instanz dieser Singleton-Klasse
	 */
	private static Coding the_instance;

	/**
	 * Erlaubt den Zugriff auf die Instanz dieser Singleton-Klasse
	 * 
	 * @return Die Instanz dieser Singleton-Klasse
	 */
	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}

	/**
	 * Gibt an, an welche Stelle in encodeDestination gerade geschrieben wird
	 * 
	 * @return Den Wert von encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * Initialisiert Coding, diese Methode muss vor einem Kodierungsvorgang aufgerufen werden. Sie setzt das Array, in das kodiert werden soll und den Zeiger auf den Anfang dieses Arrays.
	 * 
	 * @param dest Das Array, in das zukuenftig kodiert werden soll
	 */
	private void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}

	/**
	 * Initialisiert Coding zum Dekodieren. Setzt nur den Zeiger auf den Anfang des Arrays mit kodierten Daten.
	 */
	private void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * Initialisiert Coding mit dem angegebenen Array und kodiert ein kodierbares Objekt in das angegebene Array
	 * 
	 * @param o Das kodierbare Objekt
	 * @param dest Das Array, in das Kodiert werden soll
	 * @return Die Laenge des Arrays, das das kodierten Objekt enthaelt
	 */
	public short encode(Codeable o, byte[] dest) {
		encodeInit(dest);
		try {
			encode(o);
		} catch (ArrayIndexOutOfBoundsException e) {
			flowFinal();
		}
		return encoding_length;
	}

	/**
	 * Berechnet die Laenge der kodierten Version einer Klasse
	 * 
	 * @param o Das kodierbare Objekt, dessen Laenge zu berechnen ist
	 * @return Die Laenge des resultierenden Byte Arrays
	 */
	private short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	/**
	 * Kodiert eine short in das Array dieser Klasse, das kodierte Daten haelt
	 * 
	 * @param o Die zu kodierende Zahl
	 */
	private void encodeShort(short o) {
		encodeDestination[encoding_length++] = Code.SHORT;
		Util.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	private void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	/**
	 * Addiert die Laenge einer zu kodierenden short auf encode_length
	 * 
	 * @param o Die Zahl, von der die Laenge berechnet werden soll
	 */
	private void compSizeShort(short o) {
		encoding_length += 3;
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	/**
	 * Dekodiert eine short aus einem Byte-Array
	 * 
	 * @param in Das Array, aus dem dekodiert werden soll
	 * @return Die short, die im Array enthalten war
	 */
	private short decodeShort(byte[] in) {
		if (in[encoding_length++] != Code.SHORT)
			flowFinal();
		short res = Util.getShort(in, encoding_length);
		encoding_length += 2;
		return res;
	}

	private byte decodeByte(byte[] in) {
		if (in[encoding_length++] != Code.BYTE)
			flowFinal();
		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) {
		if (in[encoding_length++] != Code.BOOLEAN)
			flowFinal();
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	/**
	 * Kodiert ein byte[] nach dem TLV-Schema in das Array, das in dieser Klasse zum speichern kodierter Daten verwendet wird.
	 * 
	 * @param o Das Array, das kodiert werden soll
	 */
	private void encodeByteArray(byte[] o) {
		encodeDestination[encoding_length++] = Code.BYTEARRAY;
		Util.setShort(encodeDestination, encoding_length, (short) (o.length));
		encoding_length += 2;
		Util.arrayCopy(o, (short) 0, encodeDestination, encoding_length,
				(short) o.length);
		encoding_length += o.length;
	}

	/**
	 * Addiert die Laenge der kodierten Version eines byte[] auf encode_destination
	 * 
	 * @param o Das zu behandelnde byte[]
	 */
	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	/**
	 * Dekodiert ein byte[] aus einem Byte-Array
	 * 
	 * @param in Das Array, aus dem dekodiert werden soll
	 * @return Das byte[], die im Array enthalten war
	 */
	private void decodeByteArrayInto(byte[] in, byte[] into) {
		if (in[encoding_length++] != Code.BYTEARRAY)
			flowFinal();
		short len = Util.getShort(in, encoding_length);
		encoding_length += 2;
		Util.arrayCopy(in, encoding_length, into, (short) 0, len);
		encoding_length += len;
	}

	/**
	 * Startet den Kodierungsvorgang fuer ein kodierbares Objekt. Vor Aufruf der Methode muss encodeDestination gesetzt werden. Besser encode(Codeable c, byte[] dest) benutzen!
	 * 
	 * @param c Das zu kodierende Objekt
	 */
	private void encode(Codeable c) {
		switch (c.getCode()) {

			//Es solle eine Load kodiert werden:
			case Code.LOAD :
				encodeLoad((Load) c);
				break;

			//Es solle eine RequestBalance kodiert werden:
			case Code.REQUESTBALANCE :
				encodeRequestBalance((RequestBalance) c);
				break;

			//Es solle eine Pay kodiert werden:
			case Code.PAY :
				encodePay((Pay) c);
				break;

			//Es solle eine ResRequestBalance kodiert werden:
			case Code.RESREQUESTBALANCE :
				encodeResRequestBalance((ResRequestBalance) c);
				break;

			//Es solle eine ResPay kodiert werden:
			case Code.RESPAY :
				encodeResPay((ResPay) c);
				break;

			//Es solle eine ResAuthenticate kodiert werden:
			case Code.RESAUTHENTICATE :
				encodeResAuthenticate((ResAuthenticate) c);
				break;

			//Es solle eine Authenticate kodiert werden:
			case Code.AUTHENTICATE :
				encodeAuthenticate((Authenticate) c);
				break;

			//Es solle eine HashedData kodiert werden:
			case Code.HASHEDDATA :
				encodeHashedData((HashedData) c);
				break;

			//Es solle eine AuthData kodiert werden:
			case Code.AUTHDATA :
				encodeAuthData((AuthData) c);
				break;

			//Es solle eine Nonce kodiert werden:
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;

			//Es solle eine Secret kodiert werden:
			case Code.SECRET :
				encodeSecret((Secret) c);
				break;

			//Kein gueltiger Tag gelesen: Fehler!
			//Sollte nie erreicht werden, wird es erreicht, liegt das an der Code-Generierung! :-(
			default :
				flowFinal();
				break;
		}
	}

	public Message decodeMessage(byte[] in, short offset, short expectedLength) {
		encoding_length = offset;
		Message m = null;
		try {
			switch (in[encoding_length]) {

				//Es wird der Code einer Load gelesen:
				case Code.LOAD :
					m = Store.newLoad();
					decodeLoadInto(in, (Load) m);
					break;

				//Es wird der Code einer RequestBalance gelesen:
				case Code.REQUESTBALANCE :
					m = Store.newRequestBalance();
					decodeRequestBalanceInto(in, (RequestBalance) m);
					break;

				//Es wird der Code einer Pay gelesen:
				case Code.PAY :
					m = Store.newPay();
					decodePayInto(in, (Pay) m);
					break;

				//Es wird der Code einer ResRequestBalance gelesen:
				case Code.RESREQUESTBALANCE :
					m = Store.newResRequestBalance();
					decodeResRequestBalanceInto(in, (ResRequestBalance) m);
					break;

				//Es wird der Code einer ResPay gelesen:
				case Code.RESPAY :
					m = Store.newResPay();
					decodeResPayInto(in, (ResPay) m);
					break;

				//Es wird der Code einer ResAuthenticate gelesen:
				case Code.RESAUTHENTICATE :
					m = Store.newResAuthenticate();
					decodeResAuthenticateInto(in, (ResAuthenticate) m);
					break;

				//Es wird der Code einer Authenticate gelesen:
				case Code.AUTHENTICATE :
					m = Store.newAuthenticate();
					decodeAuthenticateInto(in, (Authenticate) m);
					break;

			}
		} catch (ArrayIndexOutOfBoundsException e) {
			flowFinal();
		}
		if (m == null || encoding_length != (short) (expectedLength + offset))
			flowFinal();
		return m;
	}

	/**
	 * Addiert die Laenge der kodierten Version einer Klasse zu encoding_length
	 *
	 * @param c Das zu behandelnde Objekt
	 */
	private void compSize(Codeable c) {
		//if (encoding_length > MAX_CODEABLE_LENGTH)
		//ISOException.throwIt(ISO7816.SW_WRONG_DATA);
		switch (c.getCode()) {

			//Es soll Load behandelt werden:
			case Code.LOAD :
				compSizeLoad((Load) c);
				break;

			//Es soll RequestBalance behandelt werden:
			case Code.REQUESTBALANCE :
				compSizeRequestBalance((RequestBalance) c);
				break;

			//Es soll Pay behandelt werden:
			case Code.PAY :
				compSizePay((Pay) c);
				break;

			//Es soll ResRequestBalance behandelt werden:
			case Code.RESREQUESTBALANCE :
				compSizeResRequestBalance((ResRequestBalance) c);
				break;

			//Es soll ResPay behandelt werden:
			case Code.RESPAY :
				compSizeResPay((ResPay) c);
				break;

			//Es soll ResAuthenticate behandelt werden:
			case Code.RESAUTHENTICATE :
				compSizeResAuthenticate((ResAuthenticate) c);
				break;

			//Es soll Authenticate behandelt werden:
			case Code.AUTHENTICATE :
				compSizeAuthenticate((Authenticate) c);
				break;

			//Es soll HashedData behandelt werden:
			case Code.HASHEDDATA :
				compSizeHashedData((HashedData) c);
				break;

			//Es soll AuthData behandelt werden:
			case Code.AUTHDATA :
				compSizeAuthData((AuthData) c);
				break;

			//Es soll Nonce behandelt werden:
			case Code.NONCE :
				compSizeNonce((Nonce) c);
				break;

			//Es soll Secret behandelt werden:
			case Code.SECRET :
				compSizeSecret((Secret) c);
				break;

			//Ein nicht behandelbares Objekt wurde uebergeben: Fehler
			//Wird dieser Code erreich, liegt das an fehlerhafter Code-Generierung :-(
			default :
				flowFinal();
				break;
		}
	}

	/**
	 * Die spezielle Kodierungsmethode fuer Load
	 *
	 * @param c Das zu kodierende Load-Objekt
	 */
	private void encodeLoad(Load c) {
		encodeDestination[encoding_length++] = Code.LOAD;

		encodeShort((short) c.amount);

		encode(c.authterminal);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer Load
	 *
	 * @param in Das byte[] in dem ein Load-Objekt kodiert ist
	 * @return Die Load-Klasse, die kodiert war
	 */
	private void decodeLoadInto(byte[] in, Load into) {
		if (in[encoding_length++] != Code.LOAD)
			flowFinal();
		//Primitiver Typ: Normale Setmethode
		into.amount = decodeShort(in);
		//Komplexer Typ:
		decodeHashedDataInto(in, (HashedData) (into.authterminal));
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts Load auf encoding_length
	 *
	 * @param c Das zu behandelnde Load-Objekt
	 */
	private void compSizeLoad(Load c) {
		encoding_length++;
		compSizeShort(c.amount);
		compSize(c.authterminal);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer RequestBalance
	 *
	 * @param c Das zu kodierende RequestBalance-Objekt
	 */
	private void encodeRequestBalance(RequestBalance c) {
		encodeDestination[encoding_length++] = Code.REQUESTBALANCE;

	}

	/**
	 * Die spezielle Dekodierungsmethode fuer RequestBalance
	 *
	 * @param in Das byte[] in dem ein RequestBalance-Objekt kodiert ist
	 * @return Die RequestBalance-Klasse, die kodiert war
	 */
	private void decodeRequestBalanceInto(byte[] in, RequestBalance into) {
		if (in[encoding_length++] != Code.REQUESTBALANCE)
			flowFinal();
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts RequestBalance auf encoding_length
	 *
	 * @param c Das zu behandelnde RequestBalance-Objekt
	 */
	private void compSizeRequestBalance(RequestBalance c) {
		encoding_length++;

	}

	/**
	 * Die spezielle Kodierungsmethode fuer Pay
	 *
	 * @param c Das zu kodierende Pay-Objekt
	 */
	private void encodePay(Pay c) {
		encodeDestination[encoding_length++] = Code.PAY;

		encodeShort((short) c.amount);

		encode(c.terminalchallenge);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer Pay
	 *
	 * @param in Das byte[] in dem ein Pay-Objekt kodiert ist
	 * @return Die Pay-Klasse, die kodiert war
	 */
	private void decodePayInto(byte[] in, Pay into) {
		if (in[encoding_length++] != Code.PAY)
			flowFinal();
		//Primitiver Typ: Normale Setmethode
		into.amount = decodeShort(in);
		//Komplexer Typ:
		decodeNonceInto(in, (Nonce) (into.terminalchallenge));
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts Pay auf encoding_length
	 *
	 * @param c Das zu behandelnde Pay-Objekt
	 */
	private void compSizePay(Pay c) {
		encoding_length++;
		compSizeShort(c.amount);
		compSize(c.terminalchallenge);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer ResRequestBalance
	 *
	 * @param c Das zu kodierende ResRequestBalance-Objekt
	 */
	private void encodeResRequestBalance(ResRequestBalance c) {
		encodeDestination[encoding_length++] = Code.RESREQUESTBALANCE;

		encodeShort((short) c.balance);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer ResRequestBalance
	 *
	 * @param in Das byte[] in dem ein ResRequestBalance-Objekt kodiert ist
	 * @return Die ResRequestBalance-Klasse, die kodiert war
	 */
	private void decodeResRequestBalanceInto(byte[] in, ResRequestBalance into) {
		if (in[encoding_length++] != Code.RESREQUESTBALANCE)
			flowFinal();
		//Primitiver Typ: Normale Setmethode
		into.balance = decodeShort(in);
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts ResRequestBalance auf encoding_length
	 *
	 * @param c Das zu behandelnde ResRequestBalance-Objekt
	 */
	private void compSizeResRequestBalance(ResRequestBalance c) {
		encoding_length++;
		compSizeShort(c.balance);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer ResPay
	 *
	 * @param c Das zu kodierende ResPay-Objekt
	 */
	private void encodeResPay(ResPay c) {
		encodeDestination[encoding_length++] = Code.RESPAY;

		encode(c.authcard);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer ResPay
	 *
	 * @param in Das byte[] in dem ein ResPay-Objekt kodiert ist
	 * @return Die ResPay-Klasse, die kodiert war
	 */
	private void decodeResPayInto(byte[] in, ResPay into) {
		if (in[encoding_length++] != Code.RESPAY)
			flowFinal();
		//Komplexer Typ:
		decodeHashedDataInto(in, (HashedData) (into.authcard));
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts ResPay auf encoding_length
	 *
	 * @param c Das zu behandelnde ResPay-Objekt
	 */
	private void compSizeResPay(ResPay c) {
		encoding_length++;
		compSize(c.authcard);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer ResAuthenticate
	 *
	 * @param c Das zu kodierende ResAuthenticate-Objekt
	 */
	private void encodeResAuthenticate(ResAuthenticate c) {
		encodeDestination[encoding_length++] = Code.RESAUTHENTICATE;

		encode(c.cardletchallenge);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer ResAuthenticate
	 *
	 * @param in Das byte[] in dem ein ResAuthenticate-Objekt kodiert ist
	 * @return Die ResAuthenticate-Klasse, die kodiert war
	 */
	private void decodeResAuthenticateInto(byte[] in, ResAuthenticate into) {
		if (in[encoding_length++] != Code.RESAUTHENTICATE)
			flowFinal();
		//Komplexer Typ:
		decodeNonceInto(in, (Nonce) (into.cardletchallenge));
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts ResAuthenticate auf encoding_length
	 *
	 * @param c Das zu behandelnde ResAuthenticate-Objekt
	 */
	private void compSizeResAuthenticate(ResAuthenticate c) {
		encoding_length++;
		compSize(c.cardletchallenge);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer Authenticate
	 *
	 * @param c Das zu kodierende Authenticate-Objekt
	 */
	private void encodeAuthenticate(Authenticate c) {
		encodeDestination[encoding_length++] = Code.AUTHENTICATE;

	}

	/**
	 * Die spezielle Dekodierungsmethode fuer Authenticate
	 *
	 * @param in Das byte[] in dem ein Authenticate-Objekt kodiert ist
	 * @return Die Authenticate-Klasse, die kodiert war
	 */
	private void decodeAuthenticateInto(byte[] in, Authenticate into) {
		if (in[encoding_length++] != Code.AUTHENTICATE)
			flowFinal();
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts Authenticate auf encoding_length
	 *
	 * @param c Das zu behandelnde Authenticate-Objekt
	 */
	private void compSizeAuthenticate(Authenticate c) {
		encoding_length++;

	}

	/**
	 * Die spezielle Kodierungsmethode fuer HashedData
	 *
	 * @param c Das zu kodierende HashedData-Objekt
	 */
	private void encodeHashedData(HashedData c) {
		encodeDestination[encoding_length++] = Code.HASHEDDATA;

		encodeByteArray((byte[]) c.hashed);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer HashedData
	 *
	 * @param in Das byte[] in dem ein HashedData-Objekt kodiert ist
	 * @return Die HashedData-Klasse, die kodiert war
	 */
	private void decodeHashedDataInto(byte[] in, HashedData into) {
		if (in[encoding_length++] != Code.HASHEDDATA)
			flowFinal();
		//Array aus primitiven Typen: Setmethode mit offset:
		decodeByteArrayInto(in, into.hashed);
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts HashedData auf encoding_length
	 *
	 * @param c Das zu behandelnde HashedData-Objekt
	 */
	private void compSizeHashedData(HashedData c) {
		encoding_length++;
		compSizeByteArray(c.hashed);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer AuthData
	 *
	 * @param c Das zu kodierende AuthData-Objekt
	 */
	private void encodeAuthData(AuthData c) {
		encodeDestination[encoding_length++] = Code.AUTHDATA;

		encodeShort((short) c.instruction);

		encode(c.passphrase);

		encode(c.challenge);

		encodeShort((short) c.amount);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer AuthData
	 *
	 * @param in Das byte[] in dem ein AuthData-Objekt kodiert ist
	 * @return Die AuthData-Klasse, die kodiert war
	 */
	private void decodeAuthDataInto(byte[] in, AuthData into) {
		if (in[encoding_length++] != Code.AUTHDATA)
			flowFinal();
		//Primitiver Typ: Normale Setmethode
		into.instruction = decodeShort(in);
		//Komplexer Typ:
		decodeSecretInto(in, (Secret) (into.passphrase));
		//Komplexer Typ:
		decodeNonceInto(in, (Nonce) (into.challenge));
		//Primitiver Typ: Normale Setmethode
		into.amount = decodeShort(in);
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts AuthData auf encoding_length
	 *
	 * @param c Das zu behandelnde AuthData-Objekt
	 */
	private void compSizeAuthData(AuthData c) {
		encoding_length++;
		compSizeShort(c.instruction);
		compSize(c.passphrase);
		compSize(c.challenge);
		compSizeShort(c.amount);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer Nonce
	 *
	 * @param c Das zu kodierende Nonce-Objekt
	 */
	private void encodeNonce(Nonce c) {
		encodeDestination[encoding_length++] = Code.NONCE;

		encodeByteArray((byte[]) c.nonce);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer Nonce
	 *
	 * @param in Das byte[] in dem ein Nonce-Objekt kodiert ist
	 * @return Die Nonce-Klasse, die kodiert war
	 */
	private void decodeNonceInto(byte[] in, Nonce into) {
		if (in[encoding_length++] != Code.NONCE)
			flowFinal();
		//Array aus primitiven Typen: Setmethode mit offset:
		decodeByteArrayInto(in, into.nonce);
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts Nonce auf encoding_length
	 *
	 * @param c Das zu behandelnde Nonce-Objekt
	 */
	private void compSizeNonce(Nonce c) {
		encoding_length++;
		compSizeByteArray(c.nonce);
	}

	/**
	 * Die spezielle Kodierungsmethode fuer Secret
	 *
	 * @param c Das zu kodierende Secret-Objekt
	 */
	private void encodeSecret(Secret c) {
		encodeDestination[encoding_length++] = Code.SECRET;

		encodeByteArray((byte[]) c.secret);
	}

	/**
	 * Die spezielle Dekodierungsmethode fuer Secret
	 *
	 * @param in Das byte[] in dem ein Secret-Objekt kodiert ist
	 * @return Die Secret-Klasse, die kodiert war
	 */
	private void decodeSecretInto(byte[] in, Secret into) {
		if (in[encoding_length++] != Code.SECRET)
			flowFinal();
		//Array aus primitiven Typen: Setmethode mit offset:
		decodeByteArrayInto(in, into.secret);
	}

	/**
	 * Addiert die Laenge der kodierten Version eines Objekts Secret auf encoding_length
	 *
	 * @param c Das zu behandelnde Secret-Objekt
	 */
	private void compSizeSecret(Secret c) {
		encoding_length++;
		compSizeByteArray(c.secret);
	}

	public void decodeCopycard(byte[] in, short offset, Copycard thisCopycard) {
		encoding_length = offset;
		if (in[encoding_length++] != Code.COPYCARD)
			flowFinal();

		if (in[encoding_length] != Code.IGNORE) {
			decodeSecretInto(in, thisCopycard.passphrase);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			thisCopycard.state = decodeByte(in);
		} else
			encoding_length++;
	}

	/**
	 * Methode zum Signalisieren eines Fehlers, der auf der Karte auftritt
	 */
	private void flowFinal() {
		ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
	}
}
